#!/usr/bin/python
import os, os.path, sys, re, copy, glob, itertools, datetime, math, random, numpy
import networkz as nx
from networkz.classes.graph import geodist_pair, get_bin_bounds, origin_coords, get_point_bin_id
import matplotlib.path as pth
import matplotlib.transforms as trs
from geopy import distance
import pdb

SFT = "safety"
DST = "distance"
LGD = "lagrangian"
OFF = "offset"
UFT = "risk"
SFT = "log_risk"

LNG = "longitude"
LAT = "latitude"

GRID_NBB_LAT=20
GRID_NBB_LNG=20

FACT_MAX = 1
TI_MAX = float("Inf")
EXTRA_EPRUNE = 1
EXTRA_GPRUNE = 1

########################################################
################# READING
def readDSGraph(nodes_file, edges_file, nodes_data, edges_data):
    graph = nx.read_edgelist(edges_file, nodetype=int, data=edges_data, create_using=nx.Graph())
    graph, bounds = nx.read_nodelist(nodes_file, nodetype=int, data=nodes_data, G=graph, check_extrema=True)
    graph.set_origin_from_bounds(bounds)
    print "Read G:\nN(G)=", len(graph), " E(G)=", graph.size()
    return graph, bounds

def readHull(hull_file):
    PREAMBLE_PATT = "^\s*(?P<src>\d+)\s+(?P<trg>\d+)\s+(?P<dclass>\d+)\s+(?P<id>\d+)\s+(?P<dist>\d+(\.\d+)?)\s+(?P<nb_points>\d+)\s*$"
    HEADER_PATT = "^point\t"
    PROFILING_PATT = "^### "
    data = {"nb_points":-1, "header": []}
    if read_profiling:
        data["profiling"] = {"overall" : {"lines": 0}}
    points = {}
    with open(hull_file) as fp:
        for line in fp:
            if data["nb_points"] == -1:
                tmp = re.match(PREAMBLE_PATT, line)
                if tmp is not None:
                    try:
                        data["pair_sf"] = ( int(tmp.group("src")), int(tmp.group("trg")), int(tmp.group("dclass")), int(tmp.group("id")))
                        data["geodist"] = float(tmp.group("dist"))
                        data["nb_points"] = int(tmp.group("nb_points"))
                    except ValueError:
                        data["nb_points"] = 0
                        continue
            elif len(data["header"]) == 0:
                if re.match(HEADER_PATT, line):
                    data["header"] = line.strip().split("\t")
            elif re.match(PROFILING_PATT, line):
                pass
            else:
                parsePointLine(line, points, data)
                
    if len(points) != data["nb_points"]:
        data["nb_points"] = -1
        points = {}
    return points, data

def parsePointLine(line, points, data):
    parts = line.strip().split("\t")
    if len(parts) == len(data["header"]):
        point_data = {}
        point_id = None
        for i, h in enumerate(data["header"]):
            try:
                if h == "point":
                    point_id = int(parts[i])
                elif h == "path":
                    tmp = map(int, parts[i].strip("[]").split(","))
                    point_data[h] = [(tmp[j], tmp[j+1]) for j in range(len(tmp)-1)]
                else:
                    point_data[h] = float(parts[i])
            except ValueError:
                parts[i] = None
        if point_id is not None:
            points[point_id] = point_data

def readSamples(samples_file):
    sampled_pairs = []
    with open(samples_file, "r") as fp:
        i = 0
        for line in fp:
            tmp = map(int, line.split())
            if len(tmp) == 3:
                sampled_pairs.append( (tmp[0], tmp[1], tmp[2], i))
                i += 1
    return sampled_pairs

def translatePoint(graph, point, SFT=SFT, path="path", to_sum=True):
    if to_sum:
        ds_l = graph.score_path(point[path], [SFT])
        point["max_"+SFT] = point[SFT]
        point[SFT] = ds_l[SFT]
    else:
        ds_l = graph.score_path_max(point[path], [SFT])
        point["sum_"+SFT] = point[SFT]
        point[SFT] = ds_l[SFT][0]

def translateHull(graph, points, SFT=SFT, path="path", to_sum=True):
    for k in points.keys():
        translatePoint(graph, points[k], SFT, path, to_sum)
    
def stopIterationFalse(points, i, j):
    return False
def stopIterationRes(points, i, j, resolution=-1):
    return j-i <= resolution
def stopIterationArea(points, i, j, height_m, width_m, area, ratio, height_j=None):
    if height_j is None:
        height_j = height_m
    else:
        ratio /= 2
    return ((points[i][height_m]-points[j][height_j])*(points[j][width_m]-points[i][width_m])) <= ratio*area
    
def addIntermediate(graph, pair_sf, points, i, j, DST, SFT, LGD, method_stop=stopIterationFalse, parameters_stop={}):
    lbd = (points[i][SFT] - points[j][SFT])/(points[i][DST] - points[j][DST])
    graph.composite_weights(LGD, [(DST, -lbd, 0), (SFT, 1, 0)])
    pred_l, distance_l = runDijkstra(graph, source=pair_sf[0], weight=LGD)
    path_l = nx.extract_path(graph, pred_l, pair_sf[0], pair_sf[1], weight=LGD)
    ds_l = graph.score_path(path_l, [SFT, DST, LGD])

    if ds_l[DST] < points[j][DST] and ds_l[SFT] < points[i][SFT]:
        idx = (j+i)/2.0
        points[idx] = ds_l
        points[idx]["path"] = path_l
        points[idx]["lbd"] = lbd

        if not method_stop(points, i, idx, **parameters_stop):
            addIntermediate(graph, pair_sf, points, i, idx, DST, SFT, LGD, method_stop, parameters_stop)
        if not method_stop(points, idx, j, **parameters_stop):
            addIntermediate(graph, pair_sf, points, idx, j, DST, SFT, LGD, method_stop, parameters_stop)

def addIntermediateMax(graph, pair_sf, points, i, j, csum, cmax, method_stop=stopIterationFalse, parameters_stop={}):
    cvmax = (points[i][SFT] + points[j]["max"])/2
    pred_l, distance_l = runDijkstra(graph, source=pair_sf[0], weight=csum, vmax=[(cmax, cvmax)])
    path_l = nx.extract_path(graph, pred_l, pair_sf[0], pair_sf[1], weight=csum)
    ds_l = graph.score_path(path_l, [csum])
    ds_lm = graph.score_path_max(path_l, [cmax])
    for k, v in ds_lm.items():
        ds_l[k] = v[0]

    # print ">>", ds_l[DST], points[j][DST], ds_l[SFT], points[i][SFT]
    if ds_l[DST] < points[j][DST] and ds_l[SFT] < points[i][SFT]:
        idx = (j+i)/2.0
        points[idx] = ds_l
        points[idx]["path"] = path_l
        points[idx]["max"] = cvmax

        if not method_stop(points, idx, j, **parameters_stop):
            addIntermediateMax(graph, pair_sf, points, idx, j, csum, cmax, method_stop, parameters_stop)
    else:
        idx = j
        # print points[idx]["max"], cvmax, " AREA:", ((points[i][csum]-points[j][csum])*(points[j]["max"]-points[i][cmax]))
        points[idx]["max"] = cvmax

    if not method_stop(points, i, idx, **parameters_stop):
        addIntermediateMax(graph, pair_sf, points, i, idx, csum, cmax, method_stop, parameters_stop)


def addIntermediateMaxOld(graph, pair_sf, points, i, j, csum, cmax, method_stop=stopIterationFalse, parameters_stop={}):
    cvmax = (points[j][SFT] + points[i][SFT])/2
    pred_l, distance_l = runDijkstra(graph, source=pair_sf[0], weight=csum, vmax=[(cmax, cvmax)])
    path_l = nx.extract_path(graph, pred_l, pair_sf[0], pair_sf[1], weight=csum)
    ds_l = graph.score_path(path_l, [csum])
    ds_lm = graph.score_path_max(path_l, [cmax])
    for k, v in ds_lm.items():
        ds_l[k] = v[0]

    if ds_l[DST] < points[j][DST] and ds_l[SFT] < points[i][SFT]:
        idx = (j+i)/2.0
        points[idx] = ds_l
        points[idx]["path"] = path_l
        points[idx]["max"] = cvmax

        if not method_stop(points, i, idx, **parameters_stop):
            addIntermediateMaxOld(graph, pair_sf, points, i, idx, csum, cmax, method_stop, parameters_stop)
        if not method_stop(points, idx, j, **parameters_stop):
            addIntermediateMaxOld(graph, pair_sf, points, idx, j, csum, cmax, method_stop, parameters_stop)


def runDijkstra(graph, source, weight, vmax=[], leq=False):
    if len(vmax) == 0:
        vmax_str = "-"
        pred_d, distance_d = nx.dijkstra_predecessor_and_distance(graph, source=source, weight=weight)
    else:
        vmax_str = ",".join(["%s:%s" % (k,v) for (k,v) in vmax])
        pred_d, distance_d = nx.dijkstra_predecessor_and_distance_with_max(graph, source=source, weight=weight, vmax=vmax, leq=leq)
    return pred_d, distance_d

def runFilterNodes(graph, method_filter, parameters_filter):
    subgraph = graph.filter_nodes(method_filter, parameters_filter)
    return subgraph

def runFilterCompartments(graph, compartments, method_select, parameters_select):
    subgraph = subgraphFromCompartments(graph, compartments, method_select, parameters_select)
    return subgraph
        
def computeLowerHull(graph, pair_sf, DST, SFT, LGD, prune=None, method_stop=None, parameters_stop={}):
    if method_stop is None:
        method_stop = stopIterationFalse
    pred_d, distance_d = runDijkstra(graph, source=pair_sf[0], weight=DST)
    path_d = nx.extract_path(graph, pred_d, pair_sf[0], pair_sf[1], weight=DST)
    ds_d = graph.score_path(path_d, [SFT, DST])
    ds_d.update({"path": path_d, "lbd": float("-Inf")})

    pred_s, distance_s = runDijkstra(graph, source=pair_sf[0], weight=SFT)
    path_s = nx.extract_path(graph, pred_s, pair_sf[0], pair_sf[1], weight=SFT)
    ds_s = graph.score_path(path_s, [SFT, DST])
    ds_s.update({"path": path_s, "lbd": 0})

    if (ds_d[SFT] != ds_s[SFT] and ds_d[DST] != ds_s[DST]):
        if prune is not None:
            m = max([ds_s[DST]]+[geodist_pair(graph.geodist_coords(u), graph.geodist_coords(pair_sf[0])) + geodist_pair(graph.geodist_coords(u), graph.geodist_coords(pair_sf[1])) for (u, v) in path_s])
            if prune["type"] == "ellipse":
                graph = runFilterNodes(graph, graph.node_filter_ellipse, {"fA": graph.geodist_coords(pair_sf[0]), 
                                                                          "fB": graph.geodist_coords(pair_sf[1]),
                                                                          "max_dist": m+EXTRA_EPRUNE})
            elif prune["type"] == "gride":
                graph = runFilterCompartments(graph, prune["compartments"],
                                              method_select=getCidsEllipseInclude,
                                              parameters_select={"grid": prune["grid"], "graph": graph,
                                                                 "fA": graph.geodist_coords(pair_sf[0]),
                                                                 "fB": graph.geodist_coords(pair_sf[1]),
                                                                 "dist": m, "path": path_s})


        points = {0: ds_d, 1: ds_s}
        i, j = 0, 1
        if method_stop is stopIterationArea:
            parameters_stop["area"] = (points[0][DST]-points[1][DST])*(points[1][SFT]-points[0][SFT])
            parameters_stop["height_m"] = SFT
            parameters_stop["width_m"] = DST
        addIntermediate(graph, pair_sf, points, i, j, DST, SFT, LGD, method_stop, parameters_stop)
    else:
        points = {0: ds_d}
    return points

def checkBounds(v, bounds):
    for (k, min_max) in bounds.items():
        if v.get(k) < min_max["min"] or v.get(k) > min_max["max"]:
            return False
    return True

def getGridCorners(grid):
    return dict([(x, get_bin_bounds(grid["bounds"][x], grid["nbbins"][x])) for x in grid["gkeys"]])

def getGridSquares(grid):
    corners = getGridCorners(grid)
    paired_corners = {}
    for (k, cs) in corners.items(): 
        paired_corners[k] = [(i, (cs[i], cs[i+1])) for i in range(len(cs)-1)]
    squares = [zip(*s) for s in itertools.product(*[paired_corners[x] for x in grid["gkeys"]])]
    return squares

def getGridCornersOrigin(grid, origin_coo):
    return {LAT: [origin_coords(origin_coo, v, 0) for v in get_bin_bounds(grid["bounds"][LAT], grid["nbbins"][LAT])],
            LNG: [origin_coords(origin_coo, v, 1) for v in get_bin_bounds(grid["bounds"][LNG], grid["nbbins"][LNG])]}

def getGridSquaresOrigin(grid, origin_coo):
    corners = getGridCornersOrigin(grid, origin_coo)
    paired_corners = {}
    for (k, cs) in corners.items(): 
        paired_corners[k] = [(i, (cs[i], cs[i+1])) for i in range(len(cs)-1)]
    squares = [zip(*s) for s in itertools.product(*[paired_corners[x] for x in grid["gkeys"]])]
    return squares

def prepareGridCompartments(graph, grid):
    nodes_map = dict([(n, graph.getNodeGridId(n, grid)) for n in graph.nodes()])
    compartments = graph.splitIntoCompartments(nodes_map)
    return compartments

def subgraphFromCompartments(graph, compartments, method_select, parameters_select={}):
    cids = method_select(**parameters_select)
    selected = listSelectedCompartments2D(cids, compartments)
    subgraph = nx.Graph()
    subgraph.graph = copy.deepcopy(graph.graph)
    subgraph.fillFromCompartments(selected, compartments)
    return subgraph

def listSelectedCompartments2D(cids, compartments):
    faces = []
    sides = []
    for cid in cids:
        if (cid, cid) in compartments:
            faces.append((cid, cid))
            for cin in compartments[(cid, cid)]["neighs"].intersection(cids):
                sides.append((cid, cin))
    return faces + sides

def getCidsEmpty(grid):
    return [list() for k in range(grid["nbbins"][grid["gkeys"][0]])]

def getCidsSquareInclude(grid, graph, pA, pB):
    xid = graph.getNodeGridId(x, grid)
    yid = graph.getNodeGridId(y, grid)
    cids = set
    for i in range(min(xid[0], yid[0]), max(xid[0], yid[0])+1):
        cids.update([(i,j) for j in range(min(xid[1], yid[1]), max(xid[1], yid[1])+1)])
    return cids

def makeEllipse(width, height, center, rtheta):
    theta = numpy.arange(0.0, 360.0, 1.0)*numpy.pi/180.0
    x = 0.5 * width * numpy.cos(theta)
    y = 0.5 * height * numpy.sin(theta)
    R = numpy.array([
        [numpy.cos(rtheta), -numpy.sin(rtheta)],
        [numpy.sin(rtheta), numpy.cos(rtheta)],
        ])
    x, y = numpy.dot(R, numpy.array([x, y]))
    x += center[0]
    y += center[1]
    return x, y

def layerSquares(squares, center_id):
    layers = {}
    for (sq, square) in squares:
        d = max([numpy.abs(sq[i]-center_id[i]) for i in range(len(center_id))])
        if d not in layers:
            layers[d] = []
        layers[d].append((sq, square))
    return layers
            
def getCidsEllipseInclude(grid, graph, fA, fB, dist, path=None):
    squares = getGridSquaresOrigin(grid, graph.get_origin_coords())
    ogfA = graph.get_point_origin_coords(fA)
    ogfB = graph.get_point_origin_coords(fB)
    center = ((ogfA[0] + ogfB[0])/2.0, (ogfA[1] + ogfB[1])/2.0)
    center_gridid = get_point_bin_id({LAT: (fA[0]+fB[0])/2.0, LNG: (fA[1]+fB[1])/2.0}, grid)
    layers = layerSquares(squares, center_gridid)
    two_c = geodist_pair(fA, fB)
    # two_c = math.sqrt((ogfA[0]-ogfB[0])**2 + (ogfA[1]-ogfB[1])**2)
    two_b = math.sqrt(dist**2 - two_c**2)
    ordered = sorted([ogfA, ogfB])
    rtheta = numpy.arccos(numpy.sign(ordered[1][1]-ordered[0][1])*(ordered[1][0]-ordered[0][0])/two_c)
    xe, ye = makeEllipse(dist+EXTRA_GPRUNE, two_b+EXTRA_GPRUNE, center, rtheta)
    pathe = pth.Path(zip(*[xe,ye]))

    cids = set()
    added = True
    layers_ids = sorted(layers.keys())
    while added and len(layers_ids)>0:
        added = False
        lsquares = layers[layers_ids.pop(0)]
        for sqi, square in lsquares:
            ry, rx = [(square[0][0], square[0][0], square[0][1], square[0][1]),
                      (square[1][0], square[1][1], square[1][1], square[1][0])]
            insi = False
            if pathe.intersects_bbox(trs.Bbox(zip(*[square[1], square[0]])), True):
                insi = True
                cids.add(sqi)
                added = True
    return cids
 
def prepareTree(graph, cmax, bounds=None):
    mstree = nx.minimum_spanning_tree(graph, weight=cmax)
    roots = []
    for cc in nx.connected_components(graph):
        if len(cc) > 1:
            root = None
            if bounds is not None:
                try:
                    root = sorted([(len(graph[n]),n) for n in cc if checkBounds(mstree.node[n], bounds)])[0][1]
                except IndexError:
                    root = None
            if root is None:
                root = sorted([(len(graph[n]),n) for n in cc])[0][1]
                # root = sorted([(len(graph[n]),n) for (n,v) in mstree.node.items()])[0][1]
            graph.label_tree_df(mstree, root, tree_pref=cmax)
            roots.append(root)
    return mstree, roots


def getTreePath(graph, pair_sf, cmax):
    path0 = graph.getPathToRoot(pair_sf[0], cmax)
    path1 = graph.getPathToRoot(pair_sf[1], cmax)
    while len(path0)> 0 and len(path1) > 0 and path0[-1] == path1[-1]:
        path0.pop()
        path1.pop()
    path1 = [(p[1], p[0]) for p in path1[::-1]]
    return path0 + path1

def getMinMax(graph, pair_sf, cmax):
    path0 = graph.getPathToRoot(pair_sf[0], cmax)
    path1 = graph.getPathToRoot(pair_sf[1], cmax)
    while len(path0)> 0 and len(path1) > 0 and path0[-1] == path1[-1]:
        path0.pop()
        path1.pop()
    maxv = float("-inf")
    for e in path0 + path1:
        if graph[e[0]][e[1]][cmax] > maxv:
            maxv = graph[e[0]][e[1]][cmax]
    return maxv

def computeLowerHullMax(graph, pair_sf, csum, cmax, prune=None, method_stop=None, parameters_stop={}):
    if method_stop is None:
        return computeLowerHullMaxOneEnd(graph, pair_sf, csum=csum, cmax=cmax, prune=prune)
    ### SHORTEST
    pred_d, distance_d = runDijkstra(graph, source=pair_sf[0], weight=csum)
    path_d = nx.extract_path(graph, pred_d, pair_sf[0], pair_sf[1], weight=csum)
    ds_d = graph.score_path(path_d, [csum])
    ds_dm = graph.score_path_max(path_d, [cmax])
    for k, v in ds_dm.items():
        ds_d[k] = v[0]
    ds_d.update({"path": path_d, "max": float("Inf")})

    ### SAFEST
    ovmax = getMinMax(graph, pair_sf, SFT)
    pred_s, distance_s = runDijkstra(graph, source=pair_sf[0], weight=csum, vmax=[(cmax, ovmax)], leq=True)
    path_s = nx.extract_path(graph, pred_s, pair_sf[0], pair_sf[1], weight=csum)
    ds_s = graph.score_path(path_s, [csum])
    ds_sm = graph.score_path_max(path_s, [cmax])
    for k, v in ds_sm.items():
        ds_s[k] = v[0]
    ds_s.update({"path": path_s, "max": ovmax})


    if (ds_d[SFT] != ds_s[SFT] and ds_d[DST] != ds_s[DST]):
        points = {0: ds_d, 1: ds_s}
        if prune is not None:
            m = max([ds_s[DST]]+[geodist_pair(graph.geodist_coords(u), graph.geodist_coords(pair_sf[0])) + geodist_pair(graph.geodist_coords(u), graph.geodist_coords(pair_sf[1])) for (u, v) in path_s])
            if prune["type"] == "ellipse":
                graph = runFilterNodes(graph, graph.node_filter_ellipse, {"fA": graph.geodist_coords(pair_sf[0]), 
                                                                          "fB": graph.geodist_coords(pair_sf[1]),
                                                                          "max_dist": m+EXTRA_EPRUNE})
            elif prune["type"] == "gride":
                graph = runFilterCompartments(graph, prune["compartments"],
                                              method_select=getCidsEllipseInclude,
                                              parameters_select={"grid": prune["grid"], "graph": graph,
                                                                 "fA": graph.geodist_coords(pair_sf[0]),
                                                                 "fB": graph.geodist_coords(pair_sf[1]),
                                                                 "dist": m, "path": path_s})

        
        i, j = 0, 1
        if method_stop is stopIterationArea:
            parameters_stop["area"] = (points[0][DST]-points[1][DST])*(points[1][SFT]-points[0][SFT])
            parameters_stop["height_m"] = SFT
            parameters_stop["width_m"] = DST
            parameters_stop["height_j"] = "max"

        addIntermediateMax(graph, pair_sf, points, i, j, csum, cmax, method_stop, parameters_stop)
    else:
        points = {0: ds_d}
    return points


def computeLowerHullMaxOneEnd(graph, pair_sf, csum, cmax, prune=None, method_stop=None, parameters_stop={}):
    pred_d, distance_d = runDijkstra(graph, source=pair_sf[0], weight=csum)
    path_d = nx.extract_path(graph, pred_d, pair_sf[0], pair_sf[1], weight=csum)
    points = {}
    cvmax = float("Inf")
    ti = 0

    if prune is not None:
        ### SAFEST
        ovmax = getMinMax(graph, pair_sf, cmax)
        pred_s, distance_s = runDijkstra(graph, source=pair_sf[0], weight=csum, vmax=[(cmax, ovmax)], leq=True)
        path_s = nx.extract_path(graph, pred_s, pair_sf[0], pair_sf[1], weight=csum)
        ds_s = graph.score_path(path_s, [csum])
        m = max([ds_s[DST]]+[geodist_pair(graph.geodist_coords(u), graph.geodist_coords(pair_sf[0])) + geodist_pair(graph.geodist_coords(u), graph.geodist_coords(pair_sf[1])) for (u, v) in path_s])

        if prune["type"] == "ellipse":
            graph = runFilterNodes(graph, graph.node_filter_ellipse, {"fA": graph.geodist_coords(pair_sf[0]), 
                                                                      "fB": graph.geodist_coords(pair_sf[1]),
                                                                      "max_dist": m+EXTRA_EPRUNE})
        elif prune["type"] == "gride":
            graph = runFilterCompartments(graph, prune["compartments"],
                                              method_select=getCidsEllipseInclude,
                                              parameters_select={"grid": prune["grid"], "graph": graph,
                                                                 "fA": graph.geodist_coords(pair_sf[0]),
                                                                 "fB": graph.geodist_coords(pair_sf[1]),
                                                                 "dist": m})
    
    while path_d is not None and ti < TI_MAX:
        ds_d = graph.score_path(path_d, [csum])
        ds_m = graph.score_path_max(path_d, [cmax])
        ds_d.update({"path": path_d, "max": cvmax})
        cvmax = ds_m[cmax][0]
        for k, v in ds_m.items():
            ds_d[k] = v[0]
        points[ti] = ds_d
        # print ti, ds_d["max"], ds_d[cmax] , ds_d[csum]
        pred_d, distance_d = runDijkstra(graph, source=pair_sf[0], weight=csum, vmax=[(cmax, FACT_MAX*cvmax)])
        path_d = nx.extract_path(graph, pred_d, pair_sf[0], pair_sf[1], weight=csum)
        ti += 1
    return points

def dispHull(points, vals=[], path=None):
    str_o = "point\t"
    str_o += "\t".join(vals)
    if path is not None:
        str_o += "\tnb_edges\t%s" % path
    for ki, k in enumerate(sorted(points.keys())):
        str_o += "\n%d\t" % ki
        str_o += "\t".join(map(str, [points[k].get(v,"---") for v in vals]))
        if path is not None:
            str_o += "\t%d" % len(points[k][path])
            str_o += "\t[" + ",".join(map(str, [p[0] for p in points[k][path]]+[points[k][path][-1][1]])) + "]"
    return str_o

def plotHull(points_sets, vals=[], steps_sets=None, title=None, legend=None, wint=None):
    import matplotlib.pyplot as plt
    import matplotlib.axes
    colors = matplotlib.axes.rcParams['axes.color_cycle']
    import pylab

    if steps_sets is not None and len(steps_sets) == len(points_sets):
        lstyle_points = "None"
        for pi, steps in enumerate(steps_sets):
            toplt = [[steps[k].get(v,0) for v in vals] for ki, k in enumerate(sorted(steps.keys()))]
            plt.plot(*zip(*toplt), linestyle=":", marker="None", color=colors[pi%len(colors)])
    else:
        lstyle_points = ":"
            
    for pi, points in enumerate(points_sets):
        toplt = [[points[k].get(v,0) for v in vals] for ki, k in enumerate(sorted(points.keys()))]
        plt.plot(*zip(*toplt), linestyle=lstyle_points, marker="+", markersize=3, color=colors[pi%len(colors)])
            
    if title is not None:
        plt.title(title)
    if legend is not None:
        plt.legend(legend)
    # if wint is not None:
    #     fig = pylab.gcf()
    #     fig.canvas.set_window_title(wint)
    plt.show()

def getSteps(points, x, y, interior=True):
    ks = sorted(points.keys())
    steps = {ks[0]: {x: points[ks[0]].get(x,0), y: points[ks[0]].get(y,0)}}
             # ks[-1]: {x: points[ks[-1]].get(x,0), y: points[ks[-1]].get(y,0)}}

    for ki in range(len(ks)-1):
        if interior:
            steps[(ks[ki]+ks[ki+1])/2.0] = {x: points[ks[ki]].get(x,0), y: points[ks[ki+1]].get(y,0)}
        else:
            steps[(ks[ki]+ks[ki+1])/2.0] = {x: points[ks[ki+1]].get(x,0), y: points[ks[ki]].get(y,0)}
        steps[ks[ki+1]] = {x: points[ks[ki+1]].get(x,0), y: points[ks[ki+1]].get(y,0)}
    return steps


def computeArea(points, width, height, stepped=0):
    ks = sorted(points.keys())
    area = 0
    min_corner = {width: min(points[ks[0]][width], points[ks[-1]][width]),
                  height: min(points[ks[0]][height], points[ks[-1]][height])}
    max_corner = {width: max(points[ks[0]][width], points[ks[-1]][width]),
                  height: max(points[ks[0]][height], points[ks[-1]][height])}

    if stepped == 0:
        ratio_tri = 2.0
    else:
        ratio_tri = 1.0

    total = (max_corner[width]-min_corner[width])*(max_corner[height]-min_corner[height])

    for ik in range(len(ks)-1):
        ### Add the triangle
        if stepped >= 0:
            area += (points[ks[ik]][width]-points[ks[ik+1]][width])*(points[ks[ik+1]][height]-points[ks[ik]][height])/ratio_tri

        ### Add the bottom rectangle
        area += (points[ks[ik+1]][width]-points[ks[ik]][width])*(points[ks[ik+1]][height]-min_corner[height])
        # print "subareas", ks[ik+1], tmpT, tmpR
    return area, total

def elist_to_nlist(elist):
    return [e[0] for e in elist]+[elist[-1][1]]
def nlist_to_elist(nlist):
    return [(nlist[i], nlist[i+1]) for i in range(len(nlist)-1)]
    
def enumeratePaths(graph, current_edge, current_path, current_costs, target, bounds, spts, paths):
    count = 1
    for nextn in graph.neighbors(current_edge[1]):
        if nextn not in current_path:
            edge_dict = graph.get_edge_data(current_edge[1], nextn)
            up_costs = dict([(measure, cv + edge_dict.get(measure)) for measure, cv in current_costs.items()])
            prune = False
            for measure, bv in bounds.items():
                if bv < up_costs[measure] + spts[measure].get(nextn):
                    prune = True
                    break
            if not prune: 
                if nextn == target:
                    paths.append({"vertices": list(current_path)+[nextn], "costs": up_costs})
                else:
                    count += enumeratePaths(graph, (current_edge[1], nextn), list(current_path)+[nextn], up_costs, target, bounds, spts, paths)
    return count
       
def processLowerHull(graph, pair_sf, typ_score, data_parameters):
    if len(pair_sf) < 4:
        pair_sf = (pair_sf[0], pair_sf[1], 0, 0)
    print "Computing lower hull %s for" % typ_score, pair_sf
    ticO = datetime.datetime.now()
    errors = "No"
    if True:
    # try:
        if typ_score == "max":
            points = computeLowerHullMax(graph, pair_sf, csum=DST, cmax=SFT, prune=data_parameters["prune"],
                                         method_stop=data_parameters["method_stop"], parameters_stop=data_parameters["parameters_stop"])
        elif typ_score == "sum":
            points = computeLowerHull(graph, pair_sf, DST, SFT, LGD, prune=data_parameters["prune"],
                                      method_stop=data_parameters["method_stop"], parameters_stop=data_parameters["parameters_stop"])
        

    fout = "%s%s_lh%s%s/%d_%d-%d.txt" % (data_parameters['xps_rep'], data_parameters['city'],
                                         typ_score, data_parameters["ext_series"],
                                         pair_sf[2], pair_sf[0], pair_sf[1])
    d = os.path.dirname(fout)
    if not os.path.exists(d):
        os.makedirs(d)
        
    with open(fout, "w") as fp:    
        fp.write("%d %d %d %d %f %d\n" % (pair_sf[0], pair_sf[1], pair_sf[2], pair_sf[3], geodist_pair(graph.geodist_coords(pair_sf[0]), graph.geodist_coords(pair_sf[1])), len(points)))
        if typ_score == "max":
            fp.write(dispHull(points, [DST, SFT, "max"], "path"))
        elif typ_score == "sum":
            fp.write(dispHull(points, [DST, SFT, LGD, "lbd"], "path"))

    # plotHull(points, [DST, SFT], path=None)

def test_filter(graph, mDST, mUFT):
    # for max_dist in [0.01, 2, 5, 20, 100]:
    for max_dist in [20, 150, 2000, 5000]:

        ticO = datetime.datetime.now()
        # subgraph = graph.filter_nodes(graph.node_filter_crow_dist, {"center": graph.geodist_coords(graph.nodes()[0]), "max_dist": max_dist})
        subgraph = graph.filter_edges(graph.edge_filter_weights, {"weights": [{"name": mDST, "thres": max_dist}, {"name": mUFT, "thres": 10**-4, "fact": -1}]})
        tocO = datetime.datetime.now()
        print "AP-- N(H)=", len(subgraph), " E(H)=", subgraph.size()
        print "TIME ADD %s" % ( tocO- ticO )

        ticO = datetime.datetime.now()
        subgraph = graph.copy()
        # subgraph.filter_nodes_inplace(graph.node_filter_crow_dist, {"center": graph.geodist_coords(graph.nodes()[0]), "max_dist": max_dist})
        subgraph.filter_edges_inplace(graph.edge_filter_weights, {"weights": [{"name": mDST, "thres": max_dist}, {"name": mUFT, "thres": 10**-4, "fact": -1}]})
        tocO = datetime.datetime.now()
        print "AP-- N(H)=", len(subgraph), " E(H)=", subgraph.size()
        print "TIME INPLACE %s" % ( tocO- ticO )

def mapGraphNodes(graph):
    points = [graph.geodist_coords(n) for n in graph.nodes()]
    plt.plot(*zip(*points), linestyle="None", marker="+", markersize=3)
    plt.show()

################ MAIN
def runHull(args):
    data_parameters= { "city" : args["city"],
                       "data_rep": "../data/"+args["city"]+"/",
                       "graph_file": "edges_lc_walk.txt",
                       "nodes_file": "nodes_lc_walk.txt",
                       "xps_rep": "../xps/",
                       "start_point": int(args["start_point"]),
                       "end_point": int(args["end_point"]),
                       "typ_score": args["typ"],
                       "prune": None,
                       "method_stop": None,
                       "parameters_stop": {},
                       "ext_series": ""}

    graph, bounds = readDSGraph(data_parameters['data_rep']+data_parameters['nodes_file'], data_parameters['data_rep']+data_parameters['graph_file'], ((LNG, float),(LAT, float)), ((DST, float),(UFT, float)))
    graph.function_weights(SFT, fct=lambda x: -numpy.log(1-x.get(UFT,0)))

    exts = []
    tmp = re.match("^a(?P<ratio>\d+)$", args["approximate"])
    if tmp is not None:
        data_parameters["method_stop"]= stopIterationArea
        data_parameters["parameters_stop"]= {"ratio": int(tmp.group("ratio"))/100.0}
        exts.append(args["approximate"])

    if args["prune"] == "p":
        data_parameters["prune"]= {"type" : "ellipse"}
        exts.append("P")

    elif args["prune"] == "g":
        grid = {"gkeys": [LNG, LAT], "bounds": bounds, "nbbins": {LNG:GRID_NBB_LNG, LAT:GRID_NBB_LAT}}
        compartments = prepareGridCompartments(graph, grid)
        data_parameters["prune"]= {"type": "gride", "grid": grid, "compartments": compartments }
        exts.append("G")

    bounds_center = {}
    for b, d in bounds.items():
        bounds_center[b] = {"min": d["min"] + 0.45*(d["max"]-d["min"]),
                            "max": d["min"] + 0.55*(d["max"]-d["min"])}
    if "max" == data_parameters["typ_score"]:
        prepareTree(graph, SFT, bounds_center)

    if len(exts) > 0:
        data_parameters["ext_series"] = "-"+"".join(sorted(exts))

    processLowerHull(graph, (data_parameters["start_point"], data_parameters["end_point"]), data_parameters["typ_score"], data_parameters)


def run(args):
    ### PARSE ARGS
    default_args = {"script" : __file__}
    default_args["city"] = "Chicago"
    # default_args["city"] = "Philadelphia"
    default_args["typ"] = "sum"
    # default_args["typ"] = "max"
    default_args["approximate"] = ""
    # default_args["approximate"] = "a10"
    default_args["prune"] = ""
    # default_args["prune"] = "p"
    default_args["start_point"] = 1032493598
    default_args["end_point"] = 375654833


    arg_order = ["script", "city", "start_point", "end_point", "typ", "approximate", "prune"]
    args = {}
    for ki, k in enumerate(arg_order):
        if len(sys.argv) > ki:
            args[k] = sys.argv[ki]
        else:
            args[k] = default_args[k]
    
    print "Working on %s" % args["city"]
    print args
    runHull(args) 

if __name__ == "__main__":
    run(sys.argv)

